package main

import (
	"bytes"
	"context"
	"encoding/json"
	"fmt"
	"io/ioutil"
	"net/http"
	"net/url"
	"strings"

	"code.gitea.io/sdk/gitea"
)

type ContentsResponse struct {
	Content []byte `json:"content"` // Assume base64 "encoding" parameter in Gitea response, and use Go's auto decode
}

type TopicsResponse struct {
	Topics []string `json:"topics"`
}

type ReaddirEntry struct {
	Name   string `json:"name"`
	Path   string `json:"path"`
	Size   int    `json:"size"`
	RawURL string `json:"download_url"`
}

func (rde ReaddirEntry) isImage() bool {
	return strings.HasSuffix(rde.Name, `.png`) || strings.HasSuffix(rde.Name, `.jpg`) || strings.HasSuffix(rde.Name, `.jpeg`)
}

type MarkdownRequest struct {
	Context string
	Mode    string
	Text    string
	Wiki    bool
}

// reposPage gets a single page of the list of Git repositories in this organisation.
func (this *Application) reposPage(ctx context.Context, page, limit int) ([]*gitea.Repository, error) {
	err := this.apiSem.Acquire(ctx, 1)
	if err != nil {
		return nil, err // e.g. ctx closed
	}
	defer this.apiSem.Release(1)

	repos, _, err := this.gc.ListOrgRepos(this.cfg.Gitea.Org, gitea.ListOrgReposOptions{ListOptions: gitea.ListOptions{Page: page, PageSize: limit}})
	if err != nil {
		return nil, err
	}

	return repos, nil
}

// repos gets a list of Git repositories in this organisation. It may have to
// make multiple network requests.
func (this *Application) repos(ctx context.Context) ([]*gitea.Repository, error) {
	// Seems like gitea-1.13.0-rc1 returns 30 results by default, and supports up to a limit of 100
	// Make a much larger request

	ret := make([]*gitea.Repository, 0)

	nextPage := 1 // Counting starts at 1
	for {
		page, err := this.reposPage(ctx, nextPage, 300)
		if err != nil {
			return nil, err
		}

		if len(page) == 0 && len(ret) > 0 {
			return ret, nil // Found enough already
		}

		ret = append(ret, page...)
		nextPage += 1
	}
}

// repoFile gets a single file from the default branch of the git repository
// Usually the default branch is `master`.
func (this *Application) repoFile(ctx context.Context, repo, filename string) ([]byte, error) {
	err := this.apiSem.Acquire(ctx, 1)
	if err != nil {
		return nil, err // e.g. ctx closed
	}
	defer this.apiSem.Release(1)

	resp, _, err := this.gc.GetContents(this.cfg.Gitea.Org, repo, "", filename)
	if err != nil {
		return nil, err
	}

	return []byte(*resp.Content), nil
}

func (this *Application) filesInDirectory(ctx context.Context, repo, dir string) ([]ReaddirEntry, error) {
	err := this.apiSem.Acquire(ctx, 1)
	if err != nil {
		return nil, err // e.g. ctx closed
	}
	defer this.apiSem.Release(1)

	req, err := http.NewRequestWithContext(ctx, http.MethodGet, this.cfg.Gitea.URL+`api/v1/repos/`+url.PathEscape(this.cfg.Gitea.Org)+`/`+url.PathEscape(repo)+`/contents/`+dir, nil) // n.b. $dir param not escaped
	if err != nil {
		return nil, err
	}

	resp, err := http.DefaultClient.Do(req)
	if err != nil {
		return nil, err
	}
	defer resp.Body.Close()

	if resp.StatusCode != 200 {

		// "No files found" happens with a HTTP 500/404 error depending on Gitea version. Catch this special case
		if resp.StatusCode == 500 || resp.StatusCode == 404 {
			b, err := ioutil.ReadAll(resp.Body)
			if err != nil {
				return nil, err
			}
			if strings.Contains(string(b), `does not exist`) {
				return []ReaddirEntry{}, nil // no files found
			}
		}

		return nil, fmt.Errorf("HTTP %d", resp.StatusCode)
	}

	var ret []ReaddirEntry
	err = json.NewDecoder(resp.Body).Decode(&ret)
	if err != nil {
		return nil, err
	}

	return ret, nil
}

// imageFilesForRepo finds documentation images for the repository.
// It searches the dist/ and doc/ subdirectories.
func (this *Application) imageFilesForRepo(ctx context.Context, repo string) ([]ReaddirEntry, error) {

	ret := []ReaddirEntry{}

	for _, dirName := range []string{`dist`, `doc`} {

		files, err := this.filesInDirectory(ctx, repo, dirName)
		if err != nil {
			return nil, fmt.Errorf("readdir(%s): %w", dirName, err)
		}

		for _, f := range files {
			if f.isImage() {
				ret = append(ret, f)
			}
		}
	}

	return ret, nil
}

func (this *Application) topicsForRepo(ctx context.Context, repo string) ([]string, error) {
	err := this.apiSem.Acquire(ctx, 1)
	if err != nil {
		return nil, err // e.g. ctx closed
	}
	defer this.apiSem.Release(1)

	req, err := http.NewRequestWithContext(ctx, http.MethodGet, this.cfg.Gitea.URL+`api/v1/repos/`+url.PathEscape(this.cfg.Gitea.Org)+`/`+url.PathEscape(repo)+`/topics`, nil)
	if err != nil {
		return nil, err
	}

	resp, err := http.DefaultClient.Do(req)
	if err != nil {
		return nil, err
	}
	defer resp.Body.Close()

	if resp.StatusCode != 200 {
		return nil, fmt.Errorf("HTTP %d", resp.StatusCode)
	}

	var tr TopicsResponse
	err = json.NewDecoder(resp.Body).Decode(&tr)
	if err != nil {
		return nil, err
	}

	return tr.Topics, nil
}

// renderMarkdown calls the remote Gitea server's own markdown renderer.
func (this *Application) renderMarkdown(ctx context.Context, repoName string, body string) ([]byte, error) {
	err := this.apiSem.Acquire(ctx, 1)
	if err != nil {
		return nil, err // e.g. ctx closed
	}
	defer this.apiSem.Release(1)

	jb, err := json.Marshal(MarkdownRequest{
		Context: this.cfg.Gitea.URL + url.PathEscape(this.cfg.Gitea.Org) + `/` + url.PathEscape(repoName) + `/src/branch/master`,
		Mode:    "gfm", // magic constant - Github Flavoured Markdown
		Text:    body,
	})
	if err != nil {
		return nil, err
	}

	req, err := http.NewRequestWithContext(ctx, http.MethodPost, this.cfg.Gitea.URL+`api/v1/markdown/`, bytes.NewReader(jb))
	if err != nil {
		return nil, err
	}
	req.Header.Set(`Content-Type`, `application/json`)
	req.Header.Set(`Content-Length`, fmt.Sprintf("%d", len(jb)))

	resp, err := http.DefaultClient.Do(req)
	if err != nil {
		return nil, err
	}
	defer resp.Body.Close()

	if resp.StatusCode != 200 {
		return nil, fmt.Errorf("HTTP %d", resp.StatusCode)
	}

	return ioutil.ReadAll(resp.Body)
}

// renderMarkdownRaw calls the remote Gitea server's own markdown renderer.
func (this *Application) renderMarkdownRaw(ctx context.Context, body []byte) ([]byte, error) {
	err := this.apiSem.Acquire(ctx, 1)
	if err != nil {
		return nil, err // e.g. ctx closed
	}
	defer this.apiSem.Release(1)

	req, err := http.NewRequestWithContext(ctx, http.MethodPost, this.cfg.Gitea.URL+`api/v1/markdown/raw`, bytes.NewReader(body))
	if err != nil {
		return nil, err
	}
	req.Header.Set(`Content-Type`, `text/plain`)
	req.Header.Set(`Content-Length`, fmt.Sprintf("%d", len(body)))

	resp, err := http.DefaultClient.Do(req)
	if err != nil {
		return nil, err
	}
	defer resp.Body.Close()

	if resp.StatusCode != 200 {
		return nil, fmt.Errorf("HTTP %d", resp.StatusCode)
	}

	return ioutil.ReadAll(resp.Body)
}
